@@ -28,7 +28,7 @@ XXXX is just an autoincrementing number - largest is the latest.
 # TODO - add a simple method to view version numbers, and delete them
 # deletion will retain the sequence of course.
 """
-        
+
 import os, sys
 
 import codecs
@@ -48,7 +48,7 @@ import random
 
 URI_BASE = "info:local/"
 STORAGE_DIR = "./rdffilestore"
-SPECIAL_FILE_PREFIX = "."
+SPECIAL_FILE_PREFIX = "_"
 
 class FileStorageFactory(object):
     def get_store(self, uri_base=URI_BASE, store_dir=STORAGE_DIR, prefix=SPECIAL_FILE_PREFIX,
@@ -62,9 +62,9 @@ class FileStorageObject(object):
         self.uri = self.fs.uri_base[id]
         self.root_uri = self.fs.uri_base["%s/%s" % (id, u"ROOT")]
         self.manifest_uri = self.fs.uri_base["%s/%s" % (id, u"MANIFEST")]
-        
+
         #self.parts = self.fs.listParts(self.id)
-        
+
     def putRoot(self, rdf):
         """Store the root for this object - rdf can either be
         an RDFXML string or an RDFobject"""
@@ -73,23 +73,23 @@ class FileStorageObject(object):
         else:
             r = RDFobject(self.id).from_string(rdf)
             self.fs._store_rdfobject(self.id, u'ROOT', rdf)
-        
+
     def getRoot(self):
         """RDFObject for this object"""
         return self.fs._get_rdfobject(self.id, u'ROOT')
-        
+
     def add_part(self, partid, bytestream):
         return self.fs._put_part(self.id, partid, bytestream)
-    
-    def get_part(self, partid):
-        return self.fs._get_part(self.id, partid)
-        
+
+    def get_part(self, partid, stream=False):
+        return self.fs._get_part(self.id, partid, stream)
+
     def del_part(self, partid):
         return self.fs._del_part(self.id, partid)
-    
+
     def listParts(self):
         pass
-        
+
     def putManifest(self, rdf):
         """Store the manifest for this object - rdf can either be
         an RDFXML string or a Manifest"""
@@ -98,26 +98,28 @@ class FileStorageObject(object):
         else:
             r = Manifest().from_string(rdf)
             self.fs._store_manifest(self.id, u'MANIFEST', rdf)
-            
+
     def getManifest(self):
         """Manifest for this object"""
         return self.fs._get_manifest(self.id, u'MANIFEST', self.manifest_uri)
-        
-        
+
+
 class FileStorageClient(object):
     def __init__(self, uri_base, store_dir, prefix, shorty_length):
         self.store_dir = store_dir
-        self.uri_base = Namespace(uri_base)
+        self.uri_base = None
+        if uri_base:
+            self.uri_base = Namespace(uri_base)
         self.ids = {}
         self.id_parts = {}
         self.prefix = prefix
         self.shorty_length = shorty_length
         self._init_store()
-        
+
     def _get_id_from_dirpath(self, dirpath):
         path = self._get_path_from_dirpath(dirpath)
         return "".join(path)
-    
+
     def _get_path_from_dirpath(self, dirpath):
         head, tail = os.path.split(dirpath)
         path = [tail]
@@ -132,23 +134,27 @@ class FileStorageClient(object):
                 path.append(tail)
         path.reverse()
         return path
-    
+
     def _id_to_dirpath(self, id):
         dirpath = self.store_dir
         while id:
             dirpath = os.path.join(dirpath, id[:self.shorty_length])
             id = id[self.shorty_length:]
         return dirpath
-    
+
     def _init_store(self):
         if not os.path.exists(self.store_dir):
-            os.mkdir(self.store_dir)
-            f = open(os.path.join(self.store_dir, "pairtree_version0_1"), "w")
-            f.write("This directory conforms to Pairtree Version 0.1. Updated spec: http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html")
-            f.close()
-            f = open(os.path.join(self.store_dir, "pairtree_prefix"),"w")
-            f.write(self.uri_base)
-            f.close()
+            if self.uri_base:
+                os.mkdir(self.store_dir)
+                f = open(os.path.join(self.store_dir, "pairtree_version0_1"), "w")
+                f.write("This directory conforms to Pairtree Version 0.1. Updated spec: http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html")
+                f.close()
+                f = open(os.path.join(self.store_dir, "pairtree_prefix"),"w")
+                f.write(self.uri_base)
+                f.close()
+            else:
+                raise NotAPairtreeStoreException("""No uri_base set for a non-existent
+                                                    store - store cannot be instanciated""")
         else:
             if os.path.exists(os.path.join(self.store_dir, "pairtree_version0_1")):
                 """Seems to be a pairtree0_1 compliant 'store'"""
@@ -160,26 +166,26 @@ class FileStorageClient(object):
                     self.uri_base = Namespace(prefix)
             else:
                 raise NotAPairtreeStoreException
-            
+
         if not os.path.isdir(self.store_dir):
             raise NotAPairtreeStoreException
         # load existing objects from the peartree
-        for dirpath, dirnames, filenames in os.walk(self.store_dir):
-            leaf = False
-            for filename in [x for x in filenames if not x.startswith('pairtree')]:
-                if len(filename) > self.shorty_length:
-                    leaf = True
-            if leaf:
-                this_id = self._get_id_from_dirpath(dirpath)
-                self.ids[this_id] = dirpath
-                self.id_parts[this_id] = dirnames
-                
+#        for dirpath, dirnames, filenames in os.walk(self.store_dir):
+#            leaf = False
+#            for filename in [x for x in filenames if not x.startswith('pairtree')]:
+#                if len(filename) > self.shorty_length:
+#                    leaf = True
+#            if leaf:
+#                this_id = self._get_id_from_dirpath(dirpath)
+#                self.ids[this_id] = dirpath
+#                self.id_parts[this_id] = dirnames
+
     def _create(self, id):
         dirpath = self._id_to_dirpath(id)
         if not os.path.exists(dirpath):
             os.makedirs(dirpath)
-            os.mkdir(os.path.join(dirpath, u'%sROOT' % self.prefix))
-            os.mkdir(os.path.join(dirpath, u'%sMANIFEST' % self.prefix))
+            os.mkdir(os.path.join(dirpath, u'ROOT'))
+            os.mkdir(os.path.join(dirpath, u'MANIFEST'))
             # Setup the root object with a datestamp
             root = RDFobject()
             root.set_uri(self.uri_base[id])
@@ -188,23 +194,23 @@ class FileStorageClient(object):
             self._store_rdfobject(id, u'ROOT', root)
         else:
             raise ObjectAlreadyExistsException
-            
+
         return FileStorageObject(id, self)
-    
+
     def _get_latest_part(self, id, prefix, part_id):
-        dirpath = os.path.join(self._id_to_dirpath(id), "%s%s" % (prefix, part_id))
+        dirpath = os.path.join(self._id_to_dirpath(id), "%s" % (part_id))
         if prefix:
             xml_versions = [int(x[1:]) for x in os.listdir(dirpath) if x.startswith(self.prefix)]
         else:
             xml_versions = [int(x.split(self.prefix)[-1]) for x in os.listdir(dirpath)]
-            
+
         if xml_versions:
             return max(xml_versions), os.path.join(dirpath, "%s%s" % (prefix, max(xml_versions)))
         else:
             return 0, ""
-        
+
     def _store_rdfobject(self, id, part_id, rdfobject):
-        dirpath = os.path.join(self._id_to_dirpath(id), "%s%s" % (self.prefix, part_id))
+        dirpath = os.path.join(self._id_to_dirpath(id), "%s" % (part_id))
         if not os.path.exists(dirpath):
             os.mkdir(dirpath)
         current_version, current_path = self._get_latest_part(id, self.prefix, part_id)
@@ -217,16 +223,26 @@ class FileStorageClient(object):
 
     def _put_part(self, id, part_id, bytestream):
         dirpath = os.path.join(self._id_to_dirpath(id), "%s" % (part_id))
-        if not os.path.exists(dirpath):        
+        if not os.path.exists(dirpath):
             os.mkdir(dirpath)
         current_version, current_path = self._get_latest_part(id, "", part_id)
         version_no = current_version + 1
         f = open(os.path.join(dirpath, "%s%s%s" % (part_id, self.prefix, version_no)), "wb")
-        f.write(bytestream)
-        f.close()
+        try:
+            # Stream file-like objects in with buffered reads
+            if hasattr(bytestream, 'read'):
+                buffer_size = 1024 * 8
+                chunk = bytestream.read(buffer_size)
+                while chunk:
+                    f.write(chunk)
+                    chunk = bytestream.read(buffer_size)
+            else:
+                f.write(bytestream)
+        finally:
+            f.close()
         return version_no
 
-    def _get_part(self, id, part_id):
+    def _get_part(self, id, part_id, stream):
         dirpath = os.path.join(self._id_to_dirpath(id), "%s" % (part_id))
         if not os.path.exists(dirpath):
             raise ObjectNotFoundException
@@ -235,19 +251,22 @@ class FileStorageClient(object):
         if not os.path.exists(fn):
             raise ObjectNotFoundException
         f = open(fn, "rb")
-        bytestream = f.read()
-        f.close()
-        return bytestream
+        if stream:
+            return f
+        else:
+            bytestream = f.read()
+            f.close()
+            return bytestream
 
     def _store_manifest(self, id, part_id, manifest):
         current_version, current_path = self._get_latest_part(id, self.prefix, part_id)
         next_xml = current_version + 1
-        dirpath = os.path.join(self._id_to_dirpath(id), "%s%s" % (self.prefix, part_id))
+        dirpath = os.path.join(self._id_to_dirpath(id), "%s" % (part_id))
         f = codecs.open(os.path.join(dirpath, "%s%s" % (self.prefix, next_xml)), encoding='utf-8', mode='w')
         f.write(manifest.to_string())
         f.close()
         return next_xml
-    
+
     def _get_rdfobject(self, id, part_id):
         current_version, current_path = self._get_latest_part(id, self.prefix, part_id)
         r = RDFobject()
@@ -266,18 +285,18 @@ class FileStorageClient(object):
             m.from_string(f.read())
             f.close()
         return m
-    
-    
+
+
     def exists(self, id):
         dirpath = os.path.join(self._id_to_dirpath(id))
         return os.path.exists(dirpath)
-    
+
     def _get_new_id(self):
         id = "%.14d" % random.randint(0,99999999999999)
         while self.exists(id):
             id = "%.14d" % random.randint(0,99999999999999)
         return id
-    
+
     def getObject(self, id=None, create_if_doesnt_exist=True):
         if not id:
             id = self._get_new_id()
@@ -288,7 +307,7 @@ class FileStorageClient(object):
             return self._create(id)
         else:
             raise ObjectNotFoundException
-            
+
     def createObject(self, id):
         if self.exists(id):
             raise ObjectAlreadyExistsException
